@@ -12,7 +12,10 @@ to BambooHR API calls defined at http://www.bamboohr.com/api/documentation/.
 
 import datetime
 import requests
-import utils
+from . import utils
+from . import config
+from .utils import make_field_xml
+from os.path import basename
 
 # Python 3 basestring compatibility:
 try:
@@ -20,6 +23,13 @@ try:
 except NameError:
     # unicode is undefined: We are running Python 3
     unicode = str
+
+    # basestring is undefined: We are running Python 3
+    try:
+        basestring
+    except NameError:
+        basestring = str
+
     basestring = (str, bytes)
 else:
     # unicode is defined: We are running Python 2
@@ -32,7 +42,8 @@ class PyBambooHR(object):
     and an optional datatype argument (defaults to JSON). This class implements
     methods for basic CRUD operations for employees and more.
     """
-    def __init__(self, subdomain='', api_key='', datatype='JSON', underscore_keys=False):
+
+    def __init__(self, subdomain='', api_key='', datatype='JSON', underscore_keys=False, timeout=60, **kwargs):
         """
         Using the subdomain, __init__ initializes the base_url for our API calls.
         This method also sets up some headers for our HTTP requests as well as our authentication (API key).
@@ -42,10 +53,10 @@ class PyBambooHR(object):
         @param datatype: String of 'JSON' or 'XML'. Sets the Accept header for return type in our HTTP requests to BambooHR.
         """
         if not api_key:
-            raise ValueError('The `api_key` argument can not be empty. Please provide a valid BambooHR API key.')
+            api_key = config.api_key
 
         if not subdomain:
-            raise ValueError('The `subdomain` argument can not be empty. Please provide a valid BambooHR company subdomain.')
+            subdomain = config.subdomain
 
         # API Version
         self.api_version = 'v1'
@@ -53,6 +64,9 @@ class PyBambooHR(object):
         # Global headers
         self.headers = {}
 
+        # Requests timeout
+        self.timeout = timeout
+
         # Referred to in the documentation as [ Company ] sometimes.
         self.subdomain = subdomain
 
@@ -68,6 +82,9 @@ class PyBambooHR(object):
         # Some people will want to use underscore keys for employee data...
         self.underscore_keys = underscore_keys
 
+        # Ask BambooHR for information that is scheduled in the future
+        self.only_current = kwargs.get('only_current', False)
+
         # We are focusing on JSON for now.
         if self.datatype == 'XML':
             raise NotImplemented("Returning XML is not currently supported.")
@@ -84,6 +101,9 @@ class PyBambooHR(object):
             'json': 'application/json'
         }
 
+        # Whether or not to verify user fields. Defaults to False.
+        self.verify_fields = kwargs.get('verify_fields', False)
+
         # These can be used as a reference for available fields, also used to validate
         # fields in get_employee and to grab all available data if no fields are passed in
         # the same function.
@@ -131,6 +151,7 @@ class PyBambooHR(object):
             "payRate": ("currency", "The employee's CURRENT pay rate. ie: $8.25"),
             "payRateEffectiveDate": ("date", "The date most recent change was made."),
             "payType": ("pay_type", "The employee's CURRENT pay type. ie: 'hourly','salary','commission','exception hourly','monthly','piece rate','contract','daily'"),
+            "preferredName": ("text", "The employee's preferred name."),
             "ssn": ("ssn", "The employee's social security number"),
             "sin": ("sin", "The employee's Canadian Social Insurance Number"),
             "state": ("state", "The employee's state/province"),
@@ -146,6 +167,7 @@ class PyBambooHR(object):
             "workPhoneExtension": ("text", "The employees work phone extension (if any)"),
             "zipcode": ("text", "The employee's zipcode"),
             "photoUploaded": ("bool", "The employee has uploaded a photo"),
+            "isPhotoUploaded": ("bool", "The employee has uploaded a photo"),
             "rehireDate": ("date", "The date the employee was rehired"),
             "adpCompanyCode": ("list", ""),
             "adpFileNumber": ("text", ""),
@@ -165,6 +187,9 @@ class PyBambooHR(object):
             "benefitClassChangeReason": ("list", ""),
         }
 
+        # dicctionary with employees data
+        self.employees = {}
+
     def _format_employee_xml(self, employee):
         """
         Utility method for turning an employee dictionary into valid employee xml.
@@ -173,10 +198,10 @@ class PyBambooHR(object):
         """
         xml_fields = ''
         for key in employee:
-            if not self.employee_fields.get(key):
+            if not self.employee_fields.get(key) and self.verify_fields:
                 raise UserWarning("You passed in an invalid field")
             else:
-                xml_fields += '\t<field id="{0}">{1}</field>\n'.format(key, employee[key])
+                xml_fields += make_field_xml(key, employee[key], pre='\t', post='\n')
 
         # Really cheesy way to build XML... this should probably be replaced at some point.
         xml = "<employee>\n{}</employee>".format(xml_fields)
@@ -191,12 +216,14 @@ class PyBambooHR(object):
         """
         xml_fields = ''
         for k, v in row.iteritems():
-            xml_fields += '\t<field id="{0}">{1}</field>\n'.format(k, v)
+            xml_fields += make_field_xml(k, v, pre='\t', post='\n')
 
         xml = "<row>\n{}</row>".format(xml_fields)
         return xml
 
-    def _format_report_xml(self, fields, title='My Custom Report', report_format='pdf'):
+    def _format_report_xml(
+            self, fields, title='My Custom Report', report_format='pdf',
+            last_changed=None):
         """
         Utility method for turning an employee dictionary into valid employee xml.
 
@@ -204,10 +231,18 @@ class PyBambooHR(object):
         """
         xml_fields = ''
         for field in fields:
-            xml_fields += '\t\t<field id="{0}" />\n'.format(field)
+            xml_fields += make_field_xml(field, None, pre='\t\t', post='\n')
+
+        xml_filters = ''
+        if last_changed and isinstance(last_changed, datetime.datetime):
+            xml_filters = '''
+                <filters>
+                    <lastChanged includeNull="no">%s</lastChanged>
+                </filters>
+            ''' % last_changed.strftime('%Y-%m-%dT%H:%M:%SZ')
 
         # Really cheesy way to build XML... this should probably be replaced at some point.
-        xml = '''<report output="{0}">\n\t<title>{1}</title>\n\t<fields>\n{2}\t</fields>\n</report>'''.format(report_format, title, xml_fields)
+        xml = '''<report output="{0}">\n\t<title>{1}</title>\n\t{2}<fields>\n{3}\t</fields>\n</report>'''.format(report_format, title, xml_filters, xml_fields)
         return xml
 
     def add_employee(self, employee):
@@ -224,7 +259,7 @@ class PyBambooHR(object):
 
         xml = self._format_employee_xml(employee)
         url = self.base_url + 'employees/'
-        r = requests.post(url, data=xml, headers=self.headers, auth=(self.api_key, ''))
+        r = requests.post(url, timeout=self.timeout, data=xml, headers=self.headers, auth=(self.api_key, ''))
         r.raise_for_status()
 
         return {'url': r.headers['location'], 'id': r.headers['location'].replace(url, "")}
@@ -241,7 +276,7 @@ class PyBambooHR(object):
         employee = utils.camelcase_keys(employee)
         xml = self._format_employee_xml(employee)
         url = self.base_url + 'employees/{0}'.format(id)
-        r = requests.post(url, data=xml, headers=self.headers, auth=(self.api_key, ''))
+        r = requests.post(url, timeout=self.timeout, data=xml, headers=self.headers, auth=(self.api_key, ''))
         r.raise_for_status()
 
         return True
@@ -254,7 +289,7 @@ class PyBambooHR(object):
         @return: A list of employee dictionaries which is a list of employees in the directory.
         """
         url = self.base_url + 'employees/directory'
-        r = requests.get(url, headers=self.headers, auth=(self.api_key, ''))
+        r = requests.get(url, timeout=self.timeout, headers=self.headers, auth=(self.api_key, ''))
         r.raise_for_status()
 
         data = r.json()
@@ -279,7 +314,7 @@ class PyBambooHR(object):
 
         if field_list:
             for f in field_list:
-                if not self.employee_fields.get(f):
+                if not f.startswith('custom') and not self.employee_fields.get(f):
                     raise UserWarning("You passed in an invalid field")
                 else:
                     get_fields.append(f)
@@ -291,8 +326,13 @@ class PyBambooHR(object):
             'fields': ",".join(get_fields)
         }
 
+        if self.only_current == False:
+            payload.update({
+                'onlyCurrent': 'false'
+            })
+
         url = self.base_url + "employees/{0}".format(employee_id)
-        r = requests.get(url, headers=self.headers, params=payload, auth=(self.api_key, ''))
+        r = requests.get(url, timeout=self.timeout, headers=self.headers, params=payload, auth=(self.api_key, ''))
         r.raise_for_status()
 
         employee = r.json()
@@ -302,6 +342,93 @@ class PyBambooHR(object):
 
         return employee
 
+    def get_all_employees(self, field_list=None, disabledUsers=False, reloadEmployees=False):
+        """
+        API method for returning a dictionary of employees.
+
+        @param allUsers: Boolean flag that indicates if get all employees.
+        @param field_list: List of fields to return with the employee dictionary.
+        @param reloadEmployees: Boolean flag that indicates if get all employees again.
+        @return: Dictionary of dictionarys containing employees information.
+        """
+        if reloadEmployees or not self.employees:
+            self.employees = {}
+
+            users = {}
+            # get all employees (doesn't get whom don't have activity)
+            for u in self.get_meta_users().values():
+                users[str(u['employeeId'])] = True if u['status']=='enabled' else False
+            # get all enabled employees (included whom don't have activity, but are enabled)
+            for u in self.get_employee_directory():
+                users[u['id']] = True
+
+            # filter enabled/active employees
+            if not disabledUsers:
+                users = {k:v for k,v in users.items() if v}
+
+            # get employees data according to field_list
+            for i,uKey in enumerate(users.keys()):
+                if uKey not in self.employees.keys():
+                    self.employees[uKey] = self.get_employee(uKey, field_list=field_list)
+
+        return self.employees
+
+    def get_employee_photo(self, employee_id, photo_size='small'):
+        """
+        API method to get photo data for an employee
+
+        @param employee_id: String of the employee id.
+        @param photo_size: String of the size of photo to fetch. Default is 'small'.
+        @return A tuple whose first element is the photo data and second element
+        is the content type header.
+        """
+        url = self.base_url + "employees/{0}/photo".format(employee_id)
+        if photo_size:
+            url = url + "/{}".format(photo_size)
+        r = requests.get(url, timeout=self.timeout, headers=self.headers, auth=(self.api_key, ''))
+        r.raise_for_status()
+
+        return r.content, r.headers.get('content-type', '')
+
+    def get_employee_files(self, employee_id):
+        """
+        API method to get files data for an employee
+
+        @param employee_id: String of the employee id.
+        @return A dictionary with data about the files of the employee
+        """
+
+        url = self.base_url + "employees/{0}/files/view/".format(employee_id)
+        r = requests.get(url, timeout=self.timeout, headers=self.headers, auth=(self.api_key, ''))
+        r.raise_for_status()
+        data = utils.transform_table_data(r.content)
+
+        return data['employee']
+
+    def upload_employee_file(self, employee_id, file_path, category_id, share, override_file_name=None):
+        """
+        API method to upload a file for an employee
+
+        @param employee_id: String of the employee id.
+        @param file_path: String of the path of the file to upload
+        @param category_id: The category id to upload the file to
+        @param share: Boolean indicating if the file is shared with employee
+        @param override_file_name: String to name the file uploaded, overriding the local file name
+        """
+
+        file_name = override_file_name if override_file_name else basename(file_path)
+
+        with open(file_path, "rb") as f:
+            params = {"file": f,
+                      "fileName": (None, file_name),
+                      "category": (None, str(category_id)),
+                      "share": (None, "yes" if share else "no")}
+
+            url = self.base_url + "employees/{0}/files/".format(employee_id)
+            r = requests.post(url, timeout=self.timeout, headers=self.headers, auth=(self.api_key, ''), files=params)
+            r.raise_for_status()
+        return True
+
     def add_row(self, table_name, employee_id, row):
         """
         API method for adding a row to a table
@@ -315,7 +442,7 @@ class PyBambooHR(object):
         xml = self._format_row_xml(row)
         url = self.base_url + \
             "employees/{0}/tables/{1}/".format(employee_id, table_name)
-        r = requests.post(url, data=xml, headers=self.headers, auth=(self.api_key, ''))
+        r = requests.post(url, timeout=self.timeout, data=xml, headers=self.headers, auth=(self.api_key, ''))
         r.raise_for_status()
 
         return True
@@ -334,7 +461,7 @@ class PyBambooHR(object):
         xml = self._format_row_xml(row)
         url = self.base_url + \
             "employees/{0}/tables/{1}/{2}/".format(employee_id, table_name, row_id)
-        r = requests.post(url, data=xml, headers=self.headers, auth=(self.api_key, ''))
+        r = requests.post(url, timeout=self.timeout, data=xml, headers=self.headers, auth=(self.api_key, ''))
         r.raise_for_status()
 
         return True
@@ -357,8 +484,8 @@ class PyBambooHR(object):
             raise UserWarning("You requested an invalid report type. Valid values are: {0}".format(','.join([k for k in self.report_formats])))
 
         filter_duplicates = 'yes' if filter_duplicates else 'no'
-        url = self.base_url + "reports/{0}?format={1}&fd={2}".format(report_id, report_format, filter_duplicates)
-        r = requests.get(url, headers=self.headers, auth=(self.api_key, ''))
+        url = self.base_url + "reports/{0}?format={1}&fd={2}&onlyCurrent={3}".format(report_id, report_format, filter_duplicates, self.only_current)
+        r = requests.get(url, timeout=self.timeout, headers=self.headers, auth=(self.api_key, ''))
         r.raise_for_status()
 
         if report_format == 'json':
@@ -380,7 +507,9 @@ class PyBambooHR(object):
 
         return result
 
-    def request_custom_report(self, field_list, report_format='xls', title="My Custom Report", output_filename=None):
+    def request_custom_report(
+            self, field_list, report_format='xls', title="My Custom Report",
+            output_filename=None, last_changed=None):
         """
         API method for returning a custom report by field list.
         http://www.bamboohr.com/api/documentation/employees.php#requestCustomReport
@@ -393,17 +522,14 @@ class PyBambooHR(object):
         @param output_filename: String (optional) if a filename/location is passed, the results will be saved to disk
         @return: A result in the format specified. (Will vary depending on format requested.)
         """
-        report_formats = self.report_formats
-        report_formats.pop('json')
-
-        if report_format not in report_formats:
-            raise UserWarning("You requested an invalid report type. Valid values are: {0}".format(','.join([k for k in report_formats])))
+        if report_format not in self.report_formats:
+            raise UserWarning("You requested an invalid report type. Valid values are: {0}".format(','.join([k for k in self.report_formats])))
 
         get_fields = []
         field_list = [utils.underscore_to_camelcase(field) for field in field_list] if field_list else None
         if field_list:
             for f in field_list:
-                if not self.employee_fields.get(f):
+                if not f.startswith('custom') and not self.employee_fields.get(f):
                     raise UserWarning("You passed in an invalid field")
                 else:
                     get_fields.append(f)
@@ -411,9 +537,11 @@ class PyBambooHR(object):
             for field in self.employee_fields:
                 get_fields.append(field)
 
-        xml = self._format_report_xml(get_fields, title=title, report_format=report_format)
+        xml = self._format_report_xml(
+            get_fields, title=title, report_format=report_format,
+            last_changed=last_changed)
         url = self.base_url + "reports/custom/?format={0}".format(report_format)
-        r = requests.post(url, data=xml, headers=self.headers, auth=(self.api_key, ''))
+        r = requests.post(url, timeout=self.timeout, data=xml, headers=self.headers, auth=(self.api_key, ''))
         r.raise_for_status()
 
         if report_format == 'json':
@@ -444,12 +572,27 @@ class PyBambooHR(object):
         the values of the table's fields for a particular date, which is stored by key 'date' in the dictionary.
         """
         url = self.base_url + 'employees/{}/tables/{}'.format(employee_id, table_name)
-        r = requests.get(url, headers=self.headers, auth=(self.api_key, ''))
+        r = requests.get(url, timeout=self.timeout, headers=self.headers, auth=(self.api_key, ''))
         r.raise_for_status()
 
         return utils.transform_tabular_data(r.content)
 
-    def get_employee_changes(self, since=None):
+    def get_employee_changed_table(self, table_name='jobInfo', since=None):
+        """
+        API method to retrieve tabular data for changed employee.
+        See http://www.bamboohr.com/api/documentation/tables.php for a list of available tables.
+
+        @return A dictionary with employee ID as key and a list of dictionaries, each dictionary showing
+        the values of the table's fields for a particular date, which is stored by key 'date' in the dictionary.
+        """
+        if not isinstance(since, datetime.datetime):
+            raise ValueError("Error: since argument must be a datetime.datetime instance")
+
+        url = 'employees/changed/tables/{}'.format(table_name)
+        params = {'since': since.strftime('%Y-%m-%dT%H:%M:%SZ')}
+        return self._query(url, params)
+
+    def get_employee_changes(self, since=None, _type=None):
         """
         Returns a list of dictionaries, each with id, action, and lastChanged keys, representing
         the employee records that have changed since the datetime object passed in the since= argument.
@@ -461,10 +604,13 @@ class PyBambooHR(object):
 
         url = self.base_url + 'employees/changed/'
         params = {'since': since.strftime('%Y-%m-%dT%H:%M:%SZ')}
-        r = requests.get(url, params=params, headers=self.headers, auth=(self.api_key, ''))
+        if _type:
+            params.update({'type': _type})
+
+        r = requests.get(url, timeout=self.timeout, params=params, headers=self.headers, auth=(self.api_key, ''))
         r.raise_for_status()
 
-        return utils.transform_change_list(r.content)
+        return r.json()
 
     def get_whos_out(self, start_date=None, end_date=None):
         start_date = utils.resolve_date_argument(start_date)
@@ -476,7 +622,101 @@ class PyBambooHR(object):
             params['start'] = start_date
         if end_date:
             params['end'] = end_date
-        r = requests.get(url, headers=self.headers, auth=(self.api_key, ''))
+        r = requests.get(url, timeout=self.timeout, params=params, headers=self.headers, auth=(self.api_key, ''))
+        r.raise_for_status()
+        return r.json()
+        # return utils.transform_whos_out(r.content)
+
+    def get_time_off_requests(self, start_date=None, end_date=None, status=None, type=None, employee_id=None):
+        start_date = utils.resolve_date_argument(start_date)
+        end_date = utils.resolve_date_argument(end_date)
+
+        params = {}
+        if start_date:
+            params['start'] = start_date
+        if end_date:
+            params['end'] = end_date
+        if status:
+            params['status'] = status
+        if type:
+            params['type'] = type
+        if employee_id:
+            params['employeeId'] = employee_id
+
+        r = self._query('time_off/requests', params, raw=True)
+        return r.json()
+        # return utils.transform_time_off(r.content)
+
+    def get_meta_fields(self):
+        """
+        API method for returning a list of fields info.
+        https://www.bamboohr.com/api/documentation/metadata.php#getFields
+        Success Response: 200
+        The HTTP Content-type header will be set with the mime type for the response.
+
+        @return: list containing fields information
+        """
+        url = self.base_url + "meta/fields/"
+        r = requests.get(url, timeout=self.timeout, headers=self.headers, auth=(self.api_key, ''))
         r.raise_for_status()
 
-        return utils.transform_whos_out(r.content)
+        return r.json()
+
+    def get_meta_tables(self):
+        """
+        API method for returning a list of tables info.
+        https://www.bamboohr.com/api/documentation/metadata.php#getTables
+        Success Response: 200
+        The HTTP Content-type header will be set with the mime type for the response.
+
+        @return: list containing tables information
+        """
+
+        url = self.base_url + "meta/tables/"
+        r = requests.get(url, timeout=self.timeout, auth=(self.api_key, ''))
+        r.raise_for_status()
+        data = utils.transform_table_data(r.content)
+        self.meta_tables = data['tables']['table']
+
+        return self.meta_tables
+
+    def get_meta_lists(self):
+        """
+        API method for returning a list with lists fields info.
+        https://www.bamboohr.com/api/documentation/metadata.php#getLists
+        Success Response: 200
+        The HTTP Content-type header will be set with the mime type for the response.
+
+        @return: list containing lists fields information
+        """
+
+        url = self.base_url + "meta/lists/"
+        r = requests.get(url, timeout=self.timeout, headers=self.headers, auth=(self.api_key, ''))
+        r.raise_for_status()
+
+        return r.json()
+
+    def get_meta_users(self):
+        """
+        API method for returning a dictionary of users info.
+        https://www.bamboohr.com/api/documentation/metadata.php#getUsers
+        Success Response: 200
+        The HTTP Content-type header will be set with the mime type for the response.
+
+        @return: dictionary containing users information
+        """
+
+        url = self.base_url + "meta/users/"
+        r = requests.get(url, timeout=self.timeout, headers=self.headers, auth=(self.api_key, ''))
+        r.raise_for_status()
+
+        return r.json()
+
+    def _query(self, url, params, raw=False):
+        url = self.base_url + url
+        r = requests.get(url, timeout=self.timeout, params=params, headers=self.headers, auth=(self.api_key, ''))
+        r.raise_for_status()
+        if raw:
+            return r
+        else:
+            return r.json()
